import numpy as np
import matplotlib
matplotlib.rcParams.update({
    "pgf.texsystem": "pdflatex",
    'font.family': 'serif',
    'text.usetex': True,
    'pgf.rcfonts': False})
import matplotlib.pylab as plt
plt.close('all')
plt.rcParams['figure.figsize'] = [8, 4]
from matplotlib.ticker import AutoMinorLocator

from gpyrn import meanfield
from gpyrn import covfunc
from gpyrn import meanfunc

time = np.linspace(0, 100, 25)
y1 = 20*np.sin(2*np.pi*time / 31)
y1err = np.random.rand(25)

y2 = 25*np.sin(2*np.pi*time / 31 + 0.5*np.pi)
y2err = np.random.rand(25)

plt.figure()
plt.errorbar(time, y1, y1err, fmt='ob', markersize=7, label='y1')
plt.errorbar(time, y2, y2err, fmt='or', markersize=7, label='y2')
plt.xlabel('Time (days)')
plt.ylabel('Measurements')
plt.legend(loc='upper right', facecolor='white', framealpha=1, edgecolor='black')
plt.grid(which='major', alpha=0.5)
plt.savefig('data2.png', bbox_inches='tight')

############## 2 datasets 
gprn = meanfield.inference(1, time, y1, y1err, y2, y2err)

nodes = [covfunc.Periodic(5, 31, 0.5)]
weight = [covfunc.SquaredExponential(5, 5), covfunc.SquaredExponential(10, 10)]
means = [meanfunc.Constant(0), meanfunc.Constant(0)]
jitter = [0.5, 0.5]

elbo, m, v = gprn.ELBOcalc(nodes, weight, means, jitter, 
                           iterations=5000, mu='init', var='init')
print('ELBO =', elbo)

tstar = np.linspace(time.min(), time.max(), 1000)

a, _, _, bb = gprn.Prediction(nodes, weight, means, jitter, tstar, m, v, separate=True)

fig = plt.figure(constrained_layout=True, figsize=(7, 7))
axs = fig.subplot_mosaic( [['predictive 1', 'node'],
                           ['predictive 1', 'node'],
                           ['predictive 2', 'weight 1'],
                           ['predictive 2', 'weight 2'],],)

axs['predictive 1'].set(xlabel='', ylabel='y1')
axs['predictive 1'].errorbar(time, y1, y1err, fmt= '.k')
axs['predictive 1'].plot(tstar, a[:,0].T, '-r')
axs['predictive 1'].xaxis.set_minor_locator(AutoMinorLocator(5))
axs['predictive 1'].yaxis.set_minor_locator(AutoMinorLocator(5))
axs['predictive 1'].grid(which='major', alpha=0.5)
axs['predictive 1'].grid(which='minor', alpha=0.2)

axs['predictive 2'].set(xlabel='', ylabel='y2')
axs['predictive 2'].errorbar(time, y2, y2err, fmt= '.k')
axs['predictive 2'].plot(tstar, a[:,1].T, '-r')
axs['predictive 2'].xaxis.set_minor_locator(AutoMinorLocator(5))
axs['predictive 2'].yaxis.set_minor_locator(AutoMinorLocator(5))
axs['predictive 2'].grid(which='major', alpha=0.5)
axs['predictive 2'].grid(which='minor', alpha=0.2)

axs['weight 1'].set(xlabel='', ylabel='1st weight')
axs['weight 1'].plot(tstar, bb[1][0].T, '-b')

axs['node'].set(xlabel='', ylabel='Node')
axs['node'].plot(tstar, bb[0].T, '-b')

axs['weight 2'].set(xlabel='', ylabel='2nd weight')
axs['weight 2'].plot(tstar, bb[1][1].T, '-b')

fig.savefig('componentsPlots.png', bbox_inches='tight')




"""
Covariance functions to use on the GPRN
"""
import numpy as np

class covFunction():
    """
    Definition the covariance functions (kernels) of our GPRN, by default and
    because it simplifies my life, all kernels include a white noise term
    """
    def __init__(self, *args):
        """ Puts all kernel arguments in an array pars """
        self.pars = np.array(args, dtype=float)
        self.pars[self.pars > 1e50] = 1e50
    def __call__(self, r, t1=None, t2=None):
        """
        r = t - t'
        Not sure if this is a good approach since will make our life harder
        when defining certain non-stationary kernels, e.g linear kernel.
        """
        raise NotImplementedError

    def __repr__(self):
        """ Representation of each kernel instance """
        return "{0}({1})".format(self.__class__.__name__,
                                 ", ".join(map(str, self.pars)))

    def __add__(self, b):
        return Sum(self, b)
    def __radd__(self, b):
        return self.__add__(b)
    def __mul__(self, b):
        return Multiplication(self, b)
    def __rmul__(self, b):
        return self.__mul__(b)


class _operator(covFunction):
    """ To allow operations between two kernels """
    def __init__(self, k1, k2):
        super(_operator, self).__init__(k1, k2)
        self.k1 = k1
        self.k2 = k2
        self.kerneltype = 'complex'

    @property
    def pars(self):
        """ Parameters og the two kernels """
        return np.append(self.k1.pars, self.k2.pars)


class Sum(_operator):
    """ To allow the sum of kernels """
    def __call__(self, r):
        return self.k1(r) + self.k2(r)

    def __repr__(self):
        return "{0} + {1}".format(self.k1, self.k2)



class Multiplication(_operator):
    """ To allow the multiplication of kernels """
    def __call__(self, r):
        return self.k1(r) * self.k2(r)

    def __repr__(self):
        return "{0} * {1}".format(self.k1, self.k2)


##### Constant #################################################################
class Constant(covFunction):
    """
    This kernel returns its constant argument c with white noise

    Parameters
    ----------
    c: float
        Constant
        
    Returns
    -------
    """
    def __init__(self, c, wn):
        super(Constant, self).__init__(c, wn)
        self.tag = 'C'
        self.c = c

    def __call__(self, r):
        return self.c**2 *np.ones_like(r)


##### White Noise ##############################################################
class WhiteNoise(covFunction):
    """
    Definition of the white noise kernel.

    Parameters
    ----------
    wn: float
        White noise amplitude

    Returns
    -------
    """
    def __init__(self, wn):
        super(WhiteNoise, self).__init__(wn)
        self.tag = 'WN'
        self.wn = wn

    def __call__(self, r):
        if r[0, :].shape == r[:, 0].shape:
            return self.wn**2 *np.diag(np.diag(np.ones_like(r)))
        return self.wn**2 *np.ones_like(r)


##### Squared exponential ######################################################
class SquaredExponential(covFunction):
    """
    Squared Exponential kernel, also known as radial basis function or RBF
    kernel in other works.

    Parameterstoo-many-arguments
    ----------
    theta: float
        Amplitude
    l: float
        Length-scale

    Returns
    -------
    """
    def __init__(self, theta, l):
        super(SquaredExponential, self).__init__(theta, l)
        self.tag = 'SE'
        self.theta = theta
        self.l = l

    def __call__(self, r):
        return self.pars[0]**2 *np.exp(-0.5 * r**2/self.pars[1]**2)


##### Periodic #################################################################
class Periodic(covFunction):
    """
    Definition of the periodic kernel.

    Parameters
    ----------
    theta: float
        Amplitude
    P: float
        Period
    lp: float
        Lenght scale

    Returns
    -------
    """
    def __init__(self, theta, P, lp):
        super(Periodic, self).__init__(theta, P, lp)
        self.tag = 'P'
        self.theta = theta
        self.lp = lp
        self.P = P

    def __call__(self, r):
        return self.pars[0]**2 *\
            np.exp(-2*np.sin(np.pi*np.abs(r)/self.pars[1])**2/self.pars[2]**2)


##### Quasi Periodic ###########################################################
class QuasiPeriodic(covFunction):
    """
    This kernel is the product between the exponential sine squared kernel
    and the squared exponential kernel, commonly known as the quasi-periodic
    kernel

    Parameters
    ----------
    theta: float
        Amplitude
    le: float
        Evolutionary time scale
    P: float
        Kernel periodicity
    lp: float
        Length scale of the periodic component

    Returns
    -------
    """
    def __init__(self, theta, le, P, lp):
        super(QuasiPeriodic, self).__init__(theta, le, P, lp)
        self.tag = 'QP'
        self.theta = theta
        self.le = le
        self.P = P
        self.lp = lp
        
    def __call__(self, r):
        return self.pars[0]**2 *np.exp(-2*np.sin(np.pi*np.abs(r)/self.pars[2])**2\
                                       /self.pars[3]**2-r**2/(2*self.pars[1]**2))


##### Rational Quadratic #######################################################
class RationalQuadratic(covFunction):
    """
    Definition of the rational quadratic kernel.
    
    Parameters
    ----------
    amplitude: float
        Amplitude of the kernel
    alpha: float
        Amplitude of large and small scale variations
    l: float
        Characteristic lenght scale to define the kernel "smoothness"
    """
    def __init__(self, amplitude, alpha, l):
        super(RationalQuadratic, self).__init__(amplitude, alpha, l)
        self.amplitude = amplitude
        self.alpha = alpha
        self.l = l
        self.params_number = 3
    def __call__(self, r):
        return self.pars[0]**2 *\
            (1+0.5*r**2/(self.pars[1]*self.pars[2]**2))**(-self.pars[1])


##### RQP kernel ###############################################################
class RQP(covFunction):
    """
    Definition of the product between the exponential sine squared kernel
    and the rational quadratic kernel that we called RQP kernel.
    If I am thinking this correctly then this kernel should tend to the
    QuasiPeriodic kernel as alpha increases, although I am not sure if we can
    say that it tends to the QuasiPeriodic kernel as alpha tends to infinity.

    Parameters
    ----------
    theta: float
        Amplitude
    alpha: float
        Alpha of the rational quadratic kernel
    ell_e, ell_p: float
        Aperiodic and periodic lenght scales
    P: float
        Periodic repetitions of the kernel
        
    Returns
    -------
    """
    def __init__(self, theta, alpha, ell_e, P, ell_p):
        super(RQP, self).__init__(theta, alpha, ell_e, P, ell_p)
        self.tag = 'RQP'
        self.theta = theta
        self.alpha = alpha
        self.ell_e = ell_e
        self.P = P
        self.ell_p = ell_p

    def __call__(self, r):
        return self.theta**2 *np.exp(-2*np.sin(np.pi*np.abs(r)/self.P)**2/self.ell_p**2) \
                        *(1+r**2/(2*self.alpha*self.ell_e**2))**(-self.alpha)


##### Cosine ###################################################################
class CoSINE(covFunction):
    """
    Definition of the cosine kernel

    Parameters
    ----------
    theta: float
        Amplitude
    P: float
        Period

    Returns
    -------
    """
    def __init__(self, theta, P):
        super(CoSINE, self).__init__(theta, P)
        self.tag = 'COS'
        self.theta = theta
        self.P = P

    def __call__(self, r):
        return self.pars[0]**2 *np.cos(2*np.pi*np.abs(r)/self.pars[1])


##### Laplacian ##############################################################
class Laplacian(covFunction):
    """
    Definition of the Laplacian kernel

    Parameters
    ----------
    theta: float
        Amplitude
    l: float
        Characteristic lenght scale

    Returns
    -------
    """
    def __init__(self, theta, l):
        super(Laplacian, self).__init__(theta, l)
        self.tag = 'LAP'
        self.theta = theta
        self.l = l
        
    def __call__(self, r):
        return self.pars[0]**2 *np.exp(-np.abs(r)/self.pars[1])


##### Exponential ##############################################################
class Exponential(covFunction):
    """
    Definition of the exponential kernel

    Parameters
    ----------
    theta: float
        Amplitude
    l: float
        Characteristic lenght scale

    Returns
    -------
    """
    def __init__(self, theta, l):
        super(Exponential, self).__init__(theta, l)
        self.tag = 'EXP'
        self.theta = theta
        self.l = l

    def __call__(self, r):
        return self.pars[0]**2 * np.exp(-np.abs(r)/self.pars[1])


##### Matern 3/2 ###############################################################
class Matern32(covFunction):
    """
    Definition of the Matern 3/2 kernel. This kernel arise when setting
    v=3/2 in the matern family of kernels

    Parameters
    ----------
    theta: float
        Amplitude
    l: float
        Characteristic lenght scale

    Returns
    -------
    """
    def __init__(self, theta, l):
        super(Matern32, self).__init__(theta, l)
        self.tag = 'M32'
        self.theta = theta
        self.ell = l

    def __call__(self, r):
        return self.pars[0]**2 *(1.0+np.sqrt(3.0)*np.abs(r)/self.pars[1]) \
            *np.exp(-np.sqrt(3.0)*np.abs(r)/self.pars[1])


#### Matern 5/2 ################################################################
class Matern52(covFunction):
    """
    Definition of the Matern 5/2 kernel. This kernel arise when setting v=5/2
    in the matern family of kernels

    Parameters
    ----------
    theta: float
        Amplitude
    l: float
        Characteristic lenght scale

    Returns
    -------
    """
    def __init__(self, theta, l):
        super(Matern52, self).__init__(theta, l)
        self.tag = 'M52'
        self.theta = theta
        self.ell = l

    def __call__(self, r):
        return self.pars[0]**2 *\
            (1.0+(3*np.sqrt(5)*self.pars[1]*np.abs(r)+5*np.abs(r)**2)/(3*self.pars[1]**2))\
                *np.exp(-np.sqrt(5.0)*np.abs(r)/self.pars[1])


#### Linear ####################################################################
class Linear(covFunction):
    """
    Definition of the Linear kernel

    Parameters
    ----------
    theta: float
        Amplitude (should we even have an amplitude???)
    c: float
        Constant
        
    Returns
    -------
    """
    def __init__(self, theta, c):
        super(Linear, self).__init__(theta, c)
        self.tag = 'LIN'
        self.theta = theta
        self.c = c

    def __call__(self, r, t1, t2):
        return  (t1 - self.pars[1]) * (t2 - self.pars[1])


##### Gamma-exponential ########################################################
class GammaExp(covFunction):
    """
    Definition of the gamma-exponential kernel

    Parameters
    ----------
    theta: float
        Amplitude
    gamma: float
        Shape parameter ( 0 < gamma <= 2)
    l: float
        Lenght scale

    Returns
    -------
    """
    def __init__(self, theta, gamma, l):
        super(GammaExp, self).__init__(theta, gamma, l)
        self.tag = 'GammaExp'
        self.theta = theta
        self.gamma = gamma
        self.l = l

    def __call__(self, r):
        return self.pars[0]**2 *np.exp(-(np.abs(r)/self.pars[2])**self.pars[1])


##### Polinomial ###############################################################
class Polynomial(covFunction):
    """
    Definition of the polinomial kernel

    Parameters
    ----------
    theta: float
        Amplitude ???
    a: float
        Real value > 0
    b: foat
        Real value >= 0
    c: int
        Integer value
    wn: float
        White noise amplitude

    Returns
    -------
    """
    def __init__(self, theta, a, b, c):
        super(Polynomial, self).__init__(theta, a, b, c)
        self.tag = 'POLY'
        self.theta = theta
        self.a = a
        self.b = b
        self.c = c

    def __call__(self, r, t1, t2):
        return (self.pars[1] * t1 * t2 + self.pars[2])**self.pars[3]


##### Piecewise ################################################################
class Piecewise(covFunction):
    """
    WARNING: EXPERIMENTAL KERNEL
    
    Parameters
    ----------
    """
    def __init__(self, eta):
        super(Piecewise, self).__init__(eta)
        self.eta = eta
        self.type = 'unknown'
    def __call__(self, r):
        r = r/(0.5*self.pars[0])
        piecewise = (3*np.abs(r) +1) * (1 - np.abs(r))**3
        piecewise = np.where(np.abs(r)>1, 0, piecewise)
        return piecewise


### END



import numpy as np
import matplotlib
matplotlib.rcParams.update({
    "pgf.texsystem": "pdflatex",
    'font.family': 'serif',
    'text.usetex': True,
    'pgf.rcfonts': False})
import matplotlib.pylab as plt
plt.close('all')
plt.rcParams['figure.figsize'] = [8, 4]
from scipy import stats
from multiprocessing import Pool

import emcee
import corner

from gpyrn import meanfield
from gpyrn import covfunc
from gpyrn import meanfunc

time = np.linspace(0, 100, 25)
y1 = 20*np.sin(2*np.pi*time / 31)
y1err = np.random.rand(25)
GPRN = meanfield.inference(1, time, y1,y1err)

#Priors
n_eta1 = stats.loguniform(0.1, 50)
n_eta3 = stats.uniform(20, 40-20)
n_eta4 = stats.loguniform(0.1, 5)
w_eta1 = stats.loguniform(0.1, 50)
w_eta2 = stats.uniform(0, 100)
jitt = stats.uniform(0, 1)

def priors():
    return np.array([n_eta1.rvs(),n_eta3.rvs(), n_eta4.rvs(),
                     w_eta1.rvs(), w_eta2.rvs(), jitt.rvs()])

def logPosterior(thetas):
    n1,n3,n4, w1,w2, j = thetas
    logprior = n_eta1.logpdf(n1)
    logprior += n_eta3.logpdf(n3)
    logprior += n_eta4.logpdf(n4)
    logprior += w_eta1.logpdf(w1)
    logprior += w_eta2.logpdf(w2)
    logprior += jitt.logpdf(j)
    if np.isinf(logprior):
        return -np.inf

    nodes = [covfunc.Periodic(n1, n3, n4)]
    weight = [covfunc.SquaredExponential(w1, w2)]
    means = [meanfunc.Constant(0)]
    jitter = [j]
    elbo, m, v = GPRN.ELBOcalc(nodes, weight, means, jitter, 
                               iterations=5000, mu='init', var='init')
    logposterior = logprior + elbo
    return logposterior

ndim = priors().size 
nwalkers = 2*ndim

pool = Pool(8)
sampler = emcee.EnsembleSampler(nwalkers, ndim, logPosterior, pool=pool)
                                
p0=[priors() for i in range(nwalkers)]
sampler.run_mcmc(p0, 5000, progress=True)

#chains plot
fig, axes = plt.subplots(6, figsize=(7, 12), sharex=True)
samples = sampler.get_chain()
labels = ["$\eta_1$", "$\eta_3$", "$\eta_4$",
          "$\\theta_1$", "$\\theta_2$", "jitter"]
for i in range(ndim):
    ax = axes[i]
    ax.plot(samples[:, :, i], "k", alpha=0.3)
    ax.set_xlim(0, len(samples))
    ax.set_ylabel(labels[i])
    ax.yaxis.set_label_coords(-0.1, 0.5)
axes[-1].set_xlabel("step number");
plt.savefig('chains.png', bbox_inches='tight')
plt.close('all')

#corner plot
flat_samples = sampler.get_chain(discard=500, thin=10, flat=True)
fig = corner.corner(flat_samples,labels=labels, color="k", bins = 50,
                    quantiles=[0.16, 0.5, 0.84], smooth=True, smooth1d=True, 
                    show_titles=True, plot_density=True, plot_contours=True,
                    fill_contours=True, plot_datapoints=False)
plt.savefig('corner.png', bbox_inches='tight')
plt.close('all')


import numpy as np
import matplotlib
matplotlib.rcParams.update({
    "pgf.texsystem": "pdflatex",
    'font.family': 'serif',
    'text.usetex': True,
    'pgf.rcfonts': False})
import matplotlib.pylab as plt
plt.close('all')
plt.rcParams['figure.figsize'] = [8, 4]
from matplotlib.ticker import AutoMinorLocator

from gpyrn import meanfield
from gpyrn import covfunc
from gpyrn import meanfunc

time = np.linspace(0, 100, 25)
y1 = 20*np.sin(2*np.pi*time / 31)
y1err = np.random.rand(25)

plt.figure()
plt.errorbar(time, y1, y1err, fmt='ob', markersize=7, label='y1')
plt.xlabel('Time (days)')
plt.ylabel('Measurements')
plt.grid(which='major', alpha=0.5)
plt.savefig('data.png', bbox_inches='tight')

############## 1 dataset - 1 node
gprn = meanfield.inference(1, time, y1, y1err)

nodes = [covfunc.Periodic(15, 31, 0.5)]
weight = [covfunc.SquaredExponential(1, 1)]
means = [meanfunc.Constant(0)]
jitter = [0.5]

elbo, m, v = gprn.ELBOcalc(nodes, weight, means, jitter, 
                           iterations=5000, mu='init', var='init')
print('ELBO =', elbo)

nodes = [covfunc.Periodic(15, 31, 0.5)]
weight = [covfunc.SquaredExponential(1, 100)]
means = [meanfunc.Constant(0)]
jitter = [0.5]

elbo, m, v = gprn.ELBOcalc(nodes, weight, means, jitter, 
                           iterations=5000, mu='init', var='init')
print('ELBO =', elbo)

tstar = np.linspace(time.min(), time.max(), 1000)
mean, _, _ = gprn.Prediction(nodes, weight, means, jitter, tstar, m, v)

plt.figure()
plt.errorbar(time, y1, y1err, fmt='ob', markersize=7, label='data')
plt.plot(tstar, mean, '--k', linewidth=2, label='predictive')
plt.xlabel('Time (days)')
plt.ylabel('Measurements')
plt.legend(loc='upper right', facecolor='white', framealpha=1, edgecolor='black')
plt.grid(which='major', alpha=0.5)
plt.savefig('dataAndPrediction.png', bbox_inches='tight')




"""
Mean functions to use on the GPRN
"""
from functools import wraps
import numpy as np

__all__ = ['Constant', 'Linear', 'Parabola', 'Cubic', 'Keplerian']

def array_input(f):
    """ Decorator to provide the __call__ methods with an array """
    @wraps(f)
    def wrapped(self, t):
        t = np.atleast_1d(t)
        r = f(self, t)
        return r
    return wrapped


class MeanModel():
    """ Class for our mean functions"""
    _parsize = 0
    def __init__(self, *pars):
        self.pars = list(pars)
        #np.array(pars, dtype=float)

    def __repr__(self):
        """ Representation of each instance """
        return "{0}({1})".format(self.__class__.__name__,
                                 ", ".join(map(str, self.pars)))

    @classmethod
    def initialize(cls):
        """ Initialize instance, setting all parameters to 0. """
        return cls(*([0.]*cls._parsize))

    def __add__(self, b):
        return Sum(self, b)
    def __radd__(self, b):
        return self.__add__(b)


class Sum(MeanModel):
    """ Sum of two mean functions """
    def __init__(self, m1, m2):
        self.m1, self.m2 = m1, m2

    @property
    def _parsize(self):
        return self.m1._parsize + self.m2._parsize

    @property
    def pars(self):
        """ Sum of parameters? """
        return self.m1.pars + self.m2.pars

    def initialize(self):
        return

    def __repr__(self):
        return "{0} + {1}".format(self.m1, self.m2)

    @array_input
    def __call__(self, t):
        return self.m1(t) + self.m2(t)


##### f(x) = a #################################################################
class Constant(MeanModel):
    """  A constant offset mean function """
    _parsize = 1
    def __init__(self, c):
        super(Constant, self).__init__(c)
        self.c = c

    @array_input
    def __call__(self, t):
        return np.full(t.shape, self.c)


##### f(x) = ax + b ############################################################
class Linear(MeanModel):
    """
    A linear mean function
    m(t) = slope * t + intercept
    """
    _parsize = 2
    def __init__(self, slope, intercept):
        super(Linear, self).__init__(slope, intercept)
        self.slope, self.intercept = slope, intercept

    @array_input
    def __call__(self, t):
        tmean = t.mean()
        return self.pars[0] * (t-tmean) + self.pars[1]


##### f(x) = ax**2 + bx + c ####################################################
class Parabola(MeanModel):
    """
    A 2nd degree polynomial mean function
    m(t) = quad * t**2 + slope * t + intercept
    """
    _parsize = 3
    def __init__(self, quad, slope, intercept):
        super(Parabola, self).__init__(quad, slope, intercept)
        self.quad, self.slope, self.intercept = quad, slope, intercept

    @array_input
    def __call__(self, t):
        return np.polyval(self.pars, t)


##### f(x) = ax**3 + bx**2 + cx + d ############################################
class Cubic(MeanModel):
    """
    A 3rd degree polynomial mean function
    m(t) = cub * t**3 + quad * t**2 + slope * t + intercept
    """
    _parsize = 4
    def __init__(self, cub, quad, slope, intercept):
        super(Cubic, self).__init__(cub, quad, slope, intercept)
        self.cub = cub
        self.quad, self.slope, self.intercept = quad, slope, intercept

    @array_input
    def __call__(self, t):
        return np.polyval(self.pars, t)


##### f(x) = a**2 * sine(2*pi*t/b + c) + d #####################################
class Sine(MeanModel):
    """
        A sinusoidal mean function
        m(t) = amplitude**2 * sine( (2*pi*t/P) + phase) + displacement
    """
    _parsize = 3
    def __init__(self, amp, P, phi, D):
        super(Sine, self).__init__(amp, P, phi, D)
        self.amp, self.P, self.phi, self.D = amp, P, phi, D

    @array_input
    def __call__(self, t):
        return self.pars[0] * np.sin((2*np.pi*t/self.pars[1]) + self.pars[2]) \
                + self.pars[3]


##### f(x) = K*(e*np.cos(w+np.cos(w+nu(x))) + d ################################
class Keplerian(MeanModel):
    """
    Keplerian function adapted from radvel
    We have RV = K[cos(w+v) + e*cos(w)] + sys_vel
    P  = period in days
    e = eccentricity
    K = RV amplitude in m/s
    w = longitude of the periastron
    phi = orbital phase
    sys_vel = offset
    """
    _parsize = 6
    def __init__(self, P, K, e, w, phi, sys_vel):
        super(Keplerian, self).__init__(P, K, e, w, phi, sys_vel)
        self.P, self.K, self.e, self.w, self.phi = P, K, e, w, phi
        self.sys_vel = sys_vel

    @array_input
    def __call__(self, t):
        P, K, e, w, phi, sys_vel = self.pars
        T0 = t[0] - (P*phi)/(2.*np.pi)
        M0 = 2*np.pi*(t-T0)/P #first guess at M
        E0 = M0 + e*np.sin(M0) + 0.5*(e**2)*np.sin(2*M0)  #first guess at E
        M1 = ( E0 - e * np.sin(E0) - M0) #goes to zero when converges
        criteria = 1e-10
        convd = np.where(np.abs(M1) > criteria)[0]  # which indices have not converged
        nd = len(convd)  # number of unconverged elements
        count = 0
        while nd > 0:
            count += 1
            E = E0
            M1p = 1 - e * np.cos(E)
            M1pp = e * np.sin(E)
            M1ppp = 1 - M1p
            d1 = -M1 / M1p
            d2 = -M1 / (M1p + d1 * M1pp / 2.0)
            d3 = -M1 / (M1p + d2 * M1pp / 2.0 + d2 * d2 * M1ppp / 6.0)
            E = E + d3
            E0 = E
            M0 = E0 - e*np.sin(E0)
            M1 = ( E0 - e * np.sin( E0 ) - M0)
            convergence_criteria = np.abs(M1) > criteria
            nd = np.sum(convergence_criteria is True)
        nu = 2*np.arctan(np.sqrt((1+e)/(1-e))*np.tan(E0/2))
        RV = K*(e*np.cos(w)+np.cos(w+nu)) + sys_vel
        return RV


##### f(x) = (x - a)**(-3) + b #################################################
class CubicSun(MeanModel):
    """
    A 3rd degree mean function
    m(t) = (t - xshift)**(-3) + yshift
    """
    _parsize = 4
    def __init__(self, xshift, yshift):
        super(Cubic, self).__init__(xshift, yshift)
        self.xshift = xshift
        self.yshift = yshift

    @array_input
    def __call__(self, t):
        return (t - self.xshift)**(-3) + self.yshift


### END



import numpy as np
import matplotlib
matplotlib.rcParams.update({
    "pgf.texsystem": "pdflatex",
    'font.family': 'serif',
    'text.usetex': True,
    'pgf.rcfonts': False})
import matplotlib.pylab as plt
plt.close('all')
from matplotlib.ticker import AutoMinorLocator

from gpyrn import meanfield
from gpyrn import covfunc
from gpyrn import meanfunc

time = np.linspace(0, 100, 25)
y1 = 20*np.sin(2*np.pi*time / 31)
y1err = np.random.rand(25)
y2 = 25*np.sin(2*np.pi*time / 31 + 0.5*np.pi)
y2err = np.random.rand(25)

plt.figure()
plt.errorbar(time, y1, y1err, fmt='ob', markersize=7, label='y1')
plt.errorbar(time, y2, y2err, fmt='or', markersize=7, label='y2')
plt.xlabel('Time (days)')
plt.ylabel('Measurements')
plt.legend(loc='upper right', facecolor='white', framealpha=1, edgecolor='black')
plt.grid(which='major', alpha=0.5)
plt.savefig('data3.png', bbox_inches='tight')

gprn = meanfield.inference(2, time, y1, y1err, y2, y2err)

nodes = [covfunc.Periodic(1, 31, 0.5),  covfunc.Matern52(1, 100)]
weight = [covfunc.SquaredExponential(20, 5), covfunc.SquaredExponential(0.1, 10),
          covfunc.SquaredExponential(10, 5), covfunc.SquaredExponential(1, 10)]
means = [meanfunc.Constant(0), meanfunc.Constant(0)]
jitter = [0.5, 0.5]

elbo, m, v = gprn.ELBOcalc(nodes, weight, means, jitter, iterations=5000, mu='init', var='init')
print('ELBO =', elbo)

tstar = np.linspace(time.min(), time.max(), 1000)
a, _, _, b = gprn.newPrediction(nodes, weight, means, jitter, tstar, m, v,
                             separate=True)

fig = plt.figure(constrained_layout=True, figsize=(7, 10))
axs = fig.subplot_mosaic( [['predictive 1', 'node 1'],
                           ['predictive 1', 'node 2'],
                           ['predictive 1', 'weight 1'],
                           ['predictive 2', 'weight 2'],
                           ['predictive 2', 'weight 3'],
                           ['predictive 2', 'weight 4'],],)

axs['predictive 1'].set(xlabel='', ylabel='y1')
axs['predictive 1'].errorbar(time, y1, y1err, fmt= '.k')
axs['predictive 1'].plot(tstar, a[:,0].T, '-r')
axs['predictive 1'].xaxis.set_minor_locator(AutoMinorLocator(5))
axs['predictive 1'].yaxis.set_minor_locator(AutoMinorLocator(5))
axs['predictive 1'].grid(which='major', alpha=0.5)
axs['predictive 1'].grid(which='minor', alpha=0.2)

axs['predictive 2'].set(xlabel='', ylabel='y2')
axs['predictive 2'].errorbar(time, y2, y2err, fmt= '.k')
axs['predictive 2'].plot(tstar, a[:,1].T, '-r')
axs['predictive 2'].xaxis.set_minor_locator(AutoMinorLocator(5))
axs['predictive 2'].yaxis.set_minor_locator(AutoMinorLocator(5))
axs['predictive 2'].grid(which='major', alpha=0.5)
axs['predictive 2'].grid(which='minor', alpha=0.2)

axs['node 1'].set(xlabel='', ylabel='1st Node')
axs['node 1'].plot(tstar, b[0][0].T, '-b')
axs['node 2'].set(xlabel='', ylabel='2nd Node')
axs['node 2'].plot(tstar, b[0][1].T, '-b')
axs['weight 1'].set(xlabel='', ylabel='1st weight')
axs['weight 1'].plot(tstar, b[1][0].T, '-b')
axs['weight 2'].set(xlabel='', ylabel='2nd weight')
axs['weight 2'].plot(tstar, b[1][1].T, '-b')

axs['weight 3'].set(xlabel='', ylabel='3rd weight')
axs['weight 3'].plot(tstar, b[1][2].T, '-b')
axs['weight 4'].set(xlabel='', ylabel='4th weight')
axs['weight 4'].plot(tstar, b[1][3].T, '-b')

fig.savefig('componentsPlots2.png', bbox_inches='tight')




"""
Collection of useful functions
"""
import matplotlib.pyplot as plt
import numpy as np

from scipy.stats import invgamma
from scipy.linalg import cho_solve, cho_factor
from scipy.optimize import minimize
from random import shuffle

##### Semi amplitude calculation ##############################################
def semi_amplitude(period, Mplanet, Mstar, ecc):
    """
    Calculates the semi-amplitude (K) caused by a planet with a given
    period and mass Mplanet, around a star of mass Mstar, with a
    eccentricity ecc.

    Parameters
    ----------
    period: float
        Period in years
    Mplanet: float
        Planet's mass in Jupiter masses, tecnically is the M.sin i
    Mstar: float
        Star mass in Solar masses
    ecc: float
        Eccentricity between 0 and 1

    Returns
    -------
    float
        Semi-amplitude K
    """
    per = np.float(np.power(1/period, 1/3))
    Pmass = Mplanet / 1
    Smass = np.float(np.power(1/Mstar, 2/3))
    Ecc = 1 / np.sqrt(1 - ecc**2)
    return 28.435 * per * Pmass* Smass * Ecc


##### Keplerian function ######################################################
def keplerian(P=365, K=.1, e=0, w=np.pi, T=0, phi=None, gamma=0, t=None):
    """
    keplerian() simulates the radial velocity signal of a planet in a
    keplerian orbit around a star.

    Parameters
    ----------
    P: float
        Period in days
    K: float
        RV amplitude
    e: float
        Eccentricity
    w: float
        Longitude of the periastron
    T: float
        Zero phase
    phi: float
        Orbital phase
    gamma: float
        Constant system RV
    t: array
        Time of measurements

    Returns
    -------
    t: array
        Time of measurements
    RV: array
        RV signal generated
    """
    if t is  None:
        print()
        print('TEMPORAL ERROR, time is nowhere to be found')
        print()
    #mean anomaly
    if phi is None:
        mean_anom = [2*np.pi*(x1-T)/P  for x1 in t]
    else:
        T = t[0] - (P*phi)/(2.*np.pi)
        mean_anom = [2*np.pi*(x1-T)/P  for x1 in t]
    #eccentric anomaly -> E0=M + e*sin(M) + 0.5*(e**2)*sin(2*M)
    E0 = [x + e*np.sin(x)  + 0.5*(e**2)*np.sin(2*x) for x in mean_anom]
    #mean anomaly -> M0=E0 - e*sin(E0)
    M0 = [x - e*np.sin(x) for x in E0]
    i = 0
    while i < 1000:
        #[x + y for x, y in zip(first, second)]
        calc_aux = [x2 - y for x2, y in zip(mean_anom, M0)]
        E1 = [x3 + y/(1-e*np.cos(x3)) for x3, y in zip(E0, calc_aux)]
        M1 = [x4 - e*np.sin(x4) for x4 in E0]
        i += 1
        E0 = E1
        M0 = M1
    nu = [2*np.arctan(np.sqrt((1+e)/(1-e))*np.tan(x5/2)) for x5 in E0]
    RV = [gamma + K*(e*np.cos(w)+np.cos(w+x6)) for x6 in nu] #m/s
    return t, RV


##### Phase-folding function ##################################################
def phase_folding(t, y, yerr, period):
    """
    phase_folding() allows the phase folding (duh...) of a given data
    accordingly to a given period

    Parameters
    ----------
    t: array
        Time
    y: array
        Measurements
    yerr: array
        Measurement errors
    period: float
        Period to fold the data

    Returns
    -------
    phase: array
        Phase
    folded_y: array
        Sorted measurments according to the phase
    folded_yerr:array
        Sorted errors according to the phase
    """
    #divide the time by the period to convert to phase
    foldtimes = t / period
    #remove the whole number part of the phase
    foldtimes = foldtimes % 1
    if yerr is None:
        yerr = 0 * y
    #sort everything
    phase, folded_y, folded_yerr = zip(*sorted(zip(foldtimes, y, yerr)))
    return phase, folded_y, folded_yerr


##### truncated cauchy distribution ###########################################
def truncCauchy_rvs(loc=0, scale=1, a=-1, b=1, size=None):
    """
    Generate random samples from a truncated Cauchy distribution.

    Parameters
    ----------
    loc: int
        Location parameter of the distribution
    scale: int
        Scale parameter of the distribution
    a, b: int
        Interval [a, b] to which the distribution is to be limited

    Returns
    -------
    rvs: float
        rvs of the truncated Cauchy
    """
    ua = np.arctan((a - loc)/scale)/np.pi + 0.5
    ub = np.arctan((b - loc)/scale)/np.pi + 0.5
    U = np.random.uniform(ua, ub, size=size)
    rvs = loc + scale * np.tan(np.pi*(U - 0.5))
    return rvs


##### inverse gamma distribution ###############################################
f = lambda x, lims: \
    (np.array([invgamma(a=x[0], scale=x[1]).cdf(lims[0]) - 0.01,
               invgamma(a=x[0], scale=x[1]).sf(lims[1]) - 0.01])**2).sum()

def invGamma(lower, upper, x0=[1, 5], showit=False):
    """
    Arguments
    ---------
    lower, upper : float
        The upper and lower limits between which we want 98% of the probability
    x0 : list, length 2
        Initial guesses for the parameters of the inverse gamma (a and scale)
    showit : bool
        Make a plot
    """
    limits = [lower, upper]
    result = minimize(f, x0=x0, args=limits, method='L-BFGS-B',
                      bounds=[(0, None), (0, None)], tol=1e-10)
    a, b = result.x
    if showit:
        _, ax = plt.subplots(1, 1, constrained_layout=True)
        d = invgamma(a=a, scale=b)
        x = np.linspace(0.2*limits[0], 2*limits[1], 1000)
        ax.plot(x, d.pdf(x))
        ax.vlines(limits, 0, d.pdf(x).max())
        plt.show()
    return invgamma(a=a, scale=b)


##### log sum ##################################################################
def log_sum(log_summands):
    """ log sum operation """
    a = np.inf
    x = log_summands.copy()
    while a == np.inf or a == -np.inf or np.isnan(a):
        a = x[0] + np.log(1 + np.sum(np.exp(x[1:] - x[0])))
        shuffle(x)
    return a


##### multivariate normal ######################################################
def multivariate_normal(r, c, method='cholesky'):
    """
    Computes multivariate normal density for "residuals" vector r and
    covariance c.

    :param array r:
        1-D array of k dimensions.

    :param array c:
        2-D array or matrix of (k x k).

    :param string method:
        Method used to compute multivariate density.
        Possible values are:
            * "cholesky": uses the Cholesky decomposition of the covariance c,
              implemented in scipy.linalg.cho_factor and scipy.linalg.cho_solve.
            * "solve": uses the numpy.linalg functions solve() and slogdet().

    :return array: multivariate density at vector position r.
    """
    # Compute normalization factor used for all methods.
    kk = len(r) * np.log(2*np.pi)
    if method == 'cholesky':
        # Use Cholesky decomposition of covariance.
        cho, lower = cho_factor(c)
        alpha = cho_solve((cho, lower), r)
        return -0.5 * (kk + np.dot(r, alpha) + 2 * np.sum(np.log(np.diag(cho))))
    if method == 'solve':
        # Use slogdet and solve
        (_, d) = np.linalg.slogdet(c)
        alpha = np.linalg.solve(c, r)
        return -0.5 * (kk + np.dot(r, alpha) + d)


##### RMS ######################################################################
def rms(array):
    """ Root mean square of array
        Parameters
        ----------
        array: array
            Measurements
            
        Returns
        -------
        rms: float
            Root mean squared error
    """
    mu = np.average(array)
    rms = np.sqrt(np.sum((array - mu)**2) / array.size)
    return rms

def wrms(array, weights):
    """ Weighted root mean square of array, given weights 
        
        Parameters
        ----------
        array: array
            Measurements
        weights: array
            weights = 1 / errors**2
            To add jitter do 1 / (errors*2 + jitter**2)
            
        Returns
        -------
        rms: float
            Weighted root mean squared error
    """
    mu = np.average(array, weights=weights)
    rms = np.sqrt(np.sum(weights * (array - mu)**2) / np.sum(weights)) 
    return rms

### END



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from setuptools import setup

setup(name='gpyrn',
      version='1.0',
      description='Modelling stellar activity in radial velocity data with Gaussian processes regression networks ',
      author='João Camacho',
      author_email='joao.camacho@astro.up.pt',
      license='MIT',
      url='https://github.com/jdavidrcamacho/gpyrn',
      packages=['gpyrn'],
      install_requires=[
        'numpy',
        'scipy',
      ],
     )



import numpy as np
from scipy.linalg import cholesky, LinAlgError
from scipy.stats import multivariate_normal
from gpyrn import _gp

class inference(object):
    """ 
    Class to perform mean field variational inference for GPRNs. 
    See Nguyen & Bonilla (2013) for more information.
    
    Parameters
    ----------
    num_nodes: int
        Number of latent node functions f(x), called f hat in the article
    time: array
        Time coordinates
    *args: arrays
        The actual data (or components), it needs be given in order of data1, 
        data1error, data2, data2error, etc...
    """ 
    def  __init__(self, num_nodes, time, *args):
        #number of node functions; f(x) in Wilson et al. (2012)
        self.num_nodes = num_nodes
        self.q = num_nodes
        #array of the time
        self.time = time 
        #number of observations, N in Wilson et al. (2012)
        self.N = self.time.size
        #the data, it should be given as data1, data1error, data2, ...
        self.args = args 
        #number of outputs y(x); p in Wilson et al. (2012)
        self.p = int(len(self.args)/2)
        #total number of weights, we will have q*p weights in total
        self.qp =  self.q * self.p
        self.d = self.time.size * self.q *(self.p+1)
        #to organize the data we now join everything
        self.tt = np.tile(time, self.p) #"extended" time
        ys = []
        ystd = []
        yerrs = []
        for i,j  in enumerate(args):
            if i%2 == 0:
                ys.append(j)
                ystd.append(np.std(j))
            else:
                yerrs.append(j)
        self.ystd = np.array(ystd).reshape(self.p, 1)
        self.y = np.array(ys).reshape(self.p, self.N) #matrix p*N of outputs
        self.yerr = np.array(yerrs).reshape(self.p, self.N) #matrix p*N of errors
        self.yerr2 = self.yerr**2
        #check if the input was correct
        assert int((i+1)/2) == self.p, \
        'Given data and number of components dont match'
        
        
##### mean functions definition ###############################################
    def _mean(self, means, time=None):
        """
        Returns the values of the mean functions
        
        Parameters
        ----------
        
        Returns
        -------
        m: float
            Value of the mean
        """
        if time is None:
            N = self.time.size
            m = np.zeros_like(self.tt)
            for i, meanfun in enumerate(means):
                if meanfun is None:
                    continue
                else:
                    m[i*N : (i+1)*N] = meanfun(self.time)
        else:
            N = time.size
            tt = np.tile(time, self.p)
            m = np.zeros_like(tt)
            for i, meanfun in enumerate(means):
                if meanfun is None:
                    continue
                else:
                    m[i*N : (i+1)*N] = meanfun(time)
        return m
    
    
##### To create matrices and samples ###########################################
    def _KMatrix(self, kernel, time = None):
        """
        Returns the covariance matrix created by evaluating a given kernel 
        at inputs time. 
        For stability issues with the GPRN a 1e-6 nugget is needed.
        
        Parameters
        ----------
        
        Returns
        -------
        K: array
            Matrix of a covariance function
        """
        r = time[:, None] - time[None, :]
        K = kernel(r) + 1e-6*np.diag(np.diag(np.ones_like(r)))
        K[np.abs(K)<1e-12] = 0.
        return K
    
    
    def _tinyNuggetKMatrix(self, kernel, time = None):
        """
        To be used in Prediction()
        Returns the covariance matrix created by evaluating a given kernel 
        at inputs time with the tiniest stability nugget possible.
        
        Returns
        -------
        K: array
            Matrix of a covariance function
        """
        r = time[:, None] - time[None, :]
        K = kernel(r) + 1.25e-12*np.diag(np.diag(np.ones_like(r)))
        return K
    
    
    def _predictKMatrix(self, kernel, time):
        """
        To be used in Prediction()
        
        Parameters
        ----------
        
        Returns
        -------
        K: array
            Matrix of a covariance function
        """
        if time.size == 1:
            r = time - self.time[None, :]
        else:
            r = time[:,None] - self.time[None,:]
        K = kernel(r) 
        return K
    
    
    def _u_to_fhatW(self, u):
        """
        Given an array of values, divides it in the corresponding nodes (f hat)
        and weights (w) parts
        
        Parameters
        ----------
        u: array
        
        Returns
        -------
        f: array
            Samples of the nodes
        w: array
            Samples of the weights
        """
        f = u[:self.q * self.N].reshape((1, self.q, self.N))
        w = u[self.q * self.N:].reshape((self.p, self.q, self.N))
        return f, w
    
    
    def _cholNugget(self, matrix, maximum=1000):
        """
        Returns the cholesky decomposition to a given matrix, if it is not
        positive definite, a nugget is added to its diagonal.
        
        Parameters
        ----------
        matrix: array
            Matrix to decompose
        maximum: int
            Number of times a nugget is added.
        
        Returns
        -------
        L: array
            Matrix containing the Cholesky factor
        nugget: float
            Nugget added to the diagonal
        """
        nugget = 0 #our nugget starts as zero
        try:
            nugget += 1e-15
            L = cholesky(matrix, lower=True, overwrite_a=True)
            return L, nugget
        except LinAlgError:
            n = 0 #number of tries
            while n < maximum:
                try:
                    L = cholesky(matrix + nugget*np.identity(matrix.shape[0]),
                                 lower=True, overwrite_a=True)
                    return L, nugget
                except LinAlgError:
                    nugget *= 10.0
                finally:
                    n += 1
            raise LinAlgError("Not positive definite, even with nugget.")
            
            
    def _initMuVar(self, nodes, weights, jitter):
        a1 = [n.pars[0] for n in nodes]
        a2 = [w.pars[0] for w in weights]
        mean1, mean2 = [], []
        var1, var2 = [], []
        for _, n in enumerate(a1):
            m = [np.sqrt(np.abs(j)*n/i)*np.sign(j) for i,j in zip(a2,self.y)]
            mean1.append(np.mean(m, axis=0))
            mean2.append([np.sqrt(np.abs(j)*i/n) for i,j in zip(a2,self.y)])
            var1.append([np.mean(jitter)*np.ones_like(self.time)])
            var2.append([jitt*np.ones_like(self.time) for jitt in jitter])
        mu = np.concatenate((mean1, mean2), axis=None)
        var = np.concatenate((var1, var2), axis=None)
        return mu, var
    
    
    def _randomMuVar(self):
        mu = np.random.randn(self.d, 1)
        var = np.random.rand(self.d, 1)
        return mu, var
    
    
    def sampleIt(self, latentFunc, time=None):
        """
        Returns samples from the kernel
        
        Parameters
        ----------
        latentFunc: func
            Covariance function
        time: array
            Time array
        
        Returns
        -------
        norm: array
            Sample of K 
        """
        if time is None:
            time = self.time
        mean = np.zeros_like(time)
        K = self._tinyNuggetKMatrix(latentFunc, time)
        normal = multivariate_normal(mean, K, allow_singular=True).rvs()
        return normal
    
    
##### Mean-Field Inference functions ##########################################
    def ELBOcalc(self, nodes, weights, means, jitter, iterations = 10000,
                     mu = None, var = None):
        """
        Function to use to calculate the evidence lower bound
        
        Parameters
        ----------
        nodes: array
            Node functions 
        weights: array
            Weight functions
        means: array
            Mean functions
        jitter: array
            Jitter terms
        iterations: int
            Number of iterations 
        mu: array
            Variational means
        var: array
            Variational variances
            
        Returns
        -------
        ELBO: array
            Value of the ELBO per iteration
        mu: array
            Optimized variational means
        var: array
            Optimized variational variance (diagonal of sigma)
        """
        #initial variational parameters
        if mu == 'random' and var == 'random':
            mu, var = self._randomMuVar()
        if mu == 'init' and var == 'init':
            mu, var = self._initMuVar(nodes, weights, jitter)
            
        jitt2 = np.array(jitter)**2
        Kf = np.array([self._KMatrix(i, self.time) for i in nodes])
        Kw = np.array([self._KMatrix(j, self.time) for j in weights])
        Lf = np.array([self._cholNugget(j)[0] for j in Kf])
        Lw = np.array([self._cholNugget(j)[0] for j in Kw])
        y = np.concatenate(self.y) - self._mean(means)
        y = np.array(np.array_split(y, self.p))
        
        #To add new elbo values inside
        ELBO, _, _, _, _ = self.ELBOaux(Kf, Kw, Lf, Lw, y, jitt2, mu, var)
        elboArray = np.array([ELBO])
        iterNumber = 0
        while iterNumber < iterations:
            #Optimize mu and var analytically
            ELBO, mu, var, sigF, sigW = self.ELBOaux(Kf, Kw, Lf, Lw, y, 
                                                     jitt2, mu, var)
            elboArray = np.append(elboArray, ELBO)
            iterNumber += 1
            #Stoping criteria:
            if iterNumber > 5:
                means = np.mean(elboArray[-5:])
                criteria = np.abs(np.std(elboArray[-5:]) / means)
                if criteria < 1e-3 and criteria !=0:
                    return ELBO, mu, var
        print('\nMax iterations reached')
        return ELBO, mu, var
    
    
    def ELBOaux(self, Kf, Kw, Lf, Lw, y, jitt2, mu, var):
        """
        Evidence Lower bound to use in ELBOcalc()
        
        Parameters
        ----------
        Kf: array
            Covariance matrices of the node functions 
        Kw: array
            Covariance matrices of the weight function
        Lf: array
            Lower matrix calculated with Cholesky of Kf 
        Lw: array
            Lower matrix calculated with Cholesky of Kw
        y: array
            Measurements - means
        jitt2: array
            Squared jitter terms
        mu: array
            Variational means
        var: array
            Variational variances
            
        Returns
        -------
        ELBO: float
            Evidence lower bound
        new_mu: array
            New variational means
        new_var: array
            New variational variances
        """ 
        #to separate the variational parameters between the nodes and weights
        muF, muW = self._u_to_fhatW(mu.flatten())
        varF, varW = self._u_to_fhatW(var.flatten())
        sigmaF, muF, sigmaW, muW = self._updateSigMu(Kf, Kw, y, jitt2, 
                                                     muF, varF, muW, varW)
        #new mean and var for the nodes
        muF = muF.reshape(1, self.q, self.N)
        varF =  np.zeros_like(varF)
        for i in range(self.q):
            varF[:,i,:] = np.diag(sigmaF[i,:,:])
        #new mean and varfor the weights
        varW =  np.zeros_like(varW)
        for j in range(self.q):
            for i in range(self.p):
                varW[i,j,:] = np.diag(sigmaW[j, i,:, :])
        new_mu = np.concatenate((muF, muW))
        new_var = np.concatenate((varF, varW))
        #Entropy
        Ent = self._entropy(sigmaF, sigmaW)
        #Expected log prior
        LogP = self._expectedLogPrior(Kf, Kw, Lf, Lw, sigmaF, muF, sigmaW, muW)
        #Expected log-likelihood
        LogL = self._expectedLogLike(y, jitt2, sigmaF, muF, sigmaW, muW)
        ELBO = (LogL + LogP + Ent) /self.q  #Evidence Lower Bound
        return ELBO, new_mu, new_var, sigmaF, sigmaW
    
    
    def _updateSigMu(self, Kf, Kw, y, jitt2, muF, varF, muW, varW):
        """
        Efficient closed-form updates fot variational parameters. This
        corresponds to eqs. 16, 17, 18, and 19 of Nguyen & Bonilla (2013)
        
        Parameters
        ----------
        Kf: array
            Covariance matrices of the node functions 
        Kw: array
            Covariance matrices of the weight function
        y: array
            Measurements - means
        jitt2: array
            Squared jitter terms
        muF: array
            Initial variational mean of each node
        varF: array
            Initial variational variance of each node
        muW: array
            Initial variational mean of each weight
        varW: array
            Initial variational variance of each weight
            
        Returns
        -------
        sigma_f: array
            Updated variational covariance of each node
        mu_f: array
            Updated variational mean of each node
        sigma_w: array
            Updated variational covariance of each weight
        mu_w: array
            Updated variational mean of each weight
        """

        Kw = Kw.reshape(self.q, self.p, self.N, self.N)
        muF = np.squeeze(muF)
        #creation of Sigma_fj and mu_fj
        sigma_f, mu_f = [], []
        for j in range(self.q):
            diagFj, auxCalc = 0, 0
            for i in range(self.p):
                diagFj = diagFj + (muW[i,j,:]*muW[i,j,:]+varW[i,j,:]) \
                                                /(jitt2[i] + self.yerr2[i,:])
                sumNj = np.zeros(self.N)
                for k in range(self.q):
                    if k != j:
                        sumNj += muW[i,k,:]*muF[k,:].reshape(self.N)
                auxCalc = auxCalc + ((y[i,:]-sumNj)*muW[i,j,:])/(jitt2[i]+self.yerr2[i,:])
            CovF = np.diag(1 / diagFj) + Kf[j]
            sigF = Kf[j] - Kf[j] @ np.linalg.solve(CovF, Kf[j])
            sigma_f.append(sigF)
            mu_f.append(sigF @ auxCalc)
            muF = np.array(mu_f)
        sigma_f = np.array(sigma_f)
        mu_f = np.array(mu_f)
        #creation of Sigma_wij and mu_wij
        sigma_w, mu_w = [], np.zeros_like(muW)
        for j in range(self.q):
            for i in range(self.p):
                mu_fj = mu_f[j]
                var_fj = np.diag(sigma_f[j])
                Diag_ij = (mu_fj*mu_fj+var_fj)/(jitt2[i]+self.yerr2[i,:])
                Kw_ij = Kw[j,i,:,:]
                CovWij = np.diag(1 / Diag_ij) + Kw_ij
                sigWij = Kw_ij - Kw_ij @ np.linalg.solve(CovWij, Kw_ij)
                sigma_w.append(sigWij)
                sumNj = np.zeros(self.N)
                for k in range(self.q):
                    if k != j:
                        sumNj += mu_f[k].reshape(self.N)*np.array(muW[i,k,:])
                auxCalc = ((y[i,:]-sumNj)*mu_f[j,:])/(jitt2[i]+self.yerr2[i,:])
                muW[i,j,:] = sigWij @ auxCalc
        sigma_w = np.array(sigma_w).reshape(self.q, self.p, self.N, self.N)
        mu_w = np.array(muW)
        return sigma_f, mu_f, sigma_w, mu_w
    
    
    def _expectedLogLike(self, y, jitt2, sigma_f, mu_f, sigma_w, mu_w):
        """
        Calculates the expected log-likelihood in mean-field inference, 
        corresponds to eq.14 in Nguyen & Bonilla (2013)
        
        Parameters
        ----------
        y: array
            Measurements - means
        jitt2: array
            Squared jitter terms
        sigma_f: array 
            Variational covariance for each node
        mu_f: array
            Variational mean for each node
        sigma_w: array
            Variational covariance for each weight
        mu_w: array
            Variational mean for each weight
            
        Returns
        -------
        logl: float
            Expected log-likelihood value
        """
        logl = 0
        for p in range(self.p):
            for n in range(self.N):
                logl += np.log(2*np.pi*(jitt2[p] + self.yerr2[p,n]))
        logl *= -0.5
        sumN = []
        for n in range(self.N):
            for p in range(self.p):
                Ydiff = y[p,n] - mu_f[0,:,n] @mu_w[p,:,n].T
                bottom = jitt2[p]+self.yerr2[p,n]
                sumN.append((Ydiff.T * Ydiff)/bottom)
        logl += -0.5 * np.sum(sumN)
        value = 0
        for p in range(self.p):
            for q in range(self.q):
                value += np.sum((np.diag(sigma_f[q,:,:])*mu_w[p,q,:]*mu_w[p,q,:] +\
                                np.diag(sigma_w[q,p,:,:])*mu_f[:,q,:]*mu_f[:,q,:] +\
                                np.diag(sigma_f[q,:,:])*np.diag(sigma_w[q,p,:,:]))\
                                /(jitt2[p]+self.yerr2[p,:]))
        logl += -0.5* value
        return logl
    
    
    def _expectedLogPrior(self, Kf, Kw, Lf, Lw, sigma_f, mu_f, sigma_w, mu_w):
        """
        Calculates the expection of the log prior wrt q(f,w) in mean-field 
        inference, corresponds to eq.15 in Nguyen & Bonilla (2013)
        
        Parameters
        ----------
            Kf: array
                Covariance matrices of the node functions 
            Kw: array
                Covariance matrices of the weight function
            sigma_f: array
                Variational covariance for each node
            mu_f: array
                Variational mean for each node
            sigma_w: array
                Variational covariance for each weight
            mu_w: array
                Variational mean for each weight
        
        Returns
        -------
        logp: float
            Expected log prior value
        """
        #we have Q nodes -> j in the paper; we have P y(x)s -> i in the paper
        first_term = 0 #calculation of the first term of eq.15 of Nguyen & Bonilla (2013)
        second_term = 0 #calculation of the second term of eq.15 of Nguyen & Bonilla (2013)
        Kw = Kw.reshape(self.q, self.p, self.N, self.N)
        Lw = Lw.reshape(self.q, self.p, self.N, self.N)
        muW = mu_w.reshape(self.q, self.p, self.N)
        sumSigmaF = np.zeros_like(sigma_f[0])
        for j in range(self.q):
            Lfj = Lf[j]
            logKf = np.float(np.sum(np.log(np.diag(Lfj))))
            muK =  np.linalg.solve(Lfj, mu_f[:,j, :].reshape(self.N))
            muKmu = muK @ muK
            sumSigmaF = sumSigmaF + sigma_f[j]
            trace = np.trace(np.linalg.solve(Kf[j], sumSigmaF))
            first_term += -logKf - 0.5*(muKmu + trace)
            for i in range(self.p):
                muK = np.linalg.solve(Lw[j,i,:,:], muW[j,i]) 
                muKmu = muK @ muK
                trace = np.trace(np.linalg.solve(Kw[j,i,:,:], sigma_w[j,i,:,:]))
                second_term += -np.float(np.sum(np.log(np.diag(Lw[j,i,:,:]))))\
                                - 0.5*(muKmu + trace)
        const = -0.5*self.N*self.q*(self.p+1)*np.log(2*np.pi)
        logp = first_term + second_term + const
        return logp
    
    
    def _entropy(self, sigma_f, sigma_w):
        """
        Calculates the entropy in mean-field inference, corresponds to eq.14 
        in Nguyen & Bonilla (2013)
        
        Parameters
        ----------
            sigma_f: array
                Variational covariance for each node
            sigma_w: array
                Variational covariance for each weight
        
        Returns
        -------
        entropy: float
            Final entropy value
        """
        entropy = 0 #starts at zero then we sum everything
        for j in range(self.q):
            L1 = self._cholNugget(sigma_f[j])
            entropy += np.sum(np.log(np.diag(L1[0])))
            for i in range(self.p):
                L2 = self._cholNugget(sigma_w[j, i, :, :])
                entropy += np.sum(np.log(np.diag(L2[0])))
        const = 0.5*self.q*(self.p+1)*self.N*(1+np.log(2*np.pi))
        return entropy + const
    
    
    def Prediction(self, node, weights, means, jitters, tstar, mu, var,
                   separate = False):
        """
        Prediction for mean-field inference
        
        Parameters
        ----------
        node: array
            Node functions 
        weight: array
            Weight function
        means: array
            Mean functions
        jitter: array
            Jitter terms
        tstar: array
            Predictions time
        mu: array
            Variational means
        var: array
            Variational variances
        separate: bool
            True to return nodes and weights predictives separately

        Returns
        -------
        predictives: array
            Predicted means
        predictivesVar: array
        """
        muF, muW = self._u_to_fhatW(mu.flatten())
        varF, varW = self._u_to_fhatW(var.flatten())
        meanVal = self._mean(means, tstar)
        meanVal = np.array(np.array_split(meanVal, self.p))
        y = np.concatenate(self.y) - self._mean(means)
        y = np.array(np.array_split(y, self.p))
        weights = np.array(weights).reshape(self.q, self.p)
        jitt2 = np.array(jitters)**2
        nPred, nVar = [], []
        wPred, wVar = [], []
        for q in range(self.q):
            gpObj = _gp.GP(self.time, muF[:,q,:])
            n,nv = gpObj.prediction(node[q], tstar, muF[:,q,:].reshape(self.N), 
                                    varF[:,q,:].reshape(self.N))
            nPred.append(n)
            nVar.append(nv)
            for p in range(self.p):
                gpObj = _gp.GP(self.time, muW[p,q,:])
                w,wv = gpObj.prediction(weights[q,p], tstar, 
                                        muW[p,q,:].reshape(self.N), 
                                        varW[p,q,:].reshape(self.N))
                wPred.append(w)
                wVar.append(wv)
        nPred, nVar = np.array(nPred), np.array(nVar)
        wPredd = np.array(wPred).reshape(self.q, self.p, tstar.size) 
        wVarr = np.array(wVar).reshape(self.q, self.p, tstar.size) 
        predictives = np.zeros((tstar.size, self.p))
        predictivesVar = np.zeros((tstar.size, self.p))
        for p in range(self.p):
            predictives[:,p] += meanVal[p]
            for q in range(self.q):
                predictives[:,p] += nPred[q]*wPredd[q,p]
                predictivesVar[:,p] += wPredd[q,p]*wPredd[q,p]*nVar[q]\
                                       +wVarr[q,p]*(nVar[q] +nPred[q]*nPred[q])\
                                       +jitt2[p]
        if separate:
            predictives = np.array(predictives)
            sepPredictives = np.array([nPred,wPred], dtype=object)
            return predictives, predictivesVar, tstar, sepPredictives
        return predictives, predictivesVar, tstar
    
    
### END
